package com.jwechrs.simplegomokugame
import com.jwechrs.mathbridge.Math

//ignored by js --> import java.lang.RuntimeException

class GameController(val boardSize: Int, val winSequenceCount: Int) {
    var gameBoardStates: HashMap<ArrayList<Int>, GomokuBoard> = HashMap()
    //var boardSize: Int = 3
    //var winSequenceCount: Int = 3
    var whiteBoard: BooleanArray = BooleanArray(boardSize * boardSize)
    var blackBoard: BooleanArray = BooleanArray(boardSize * boardSize)
    var stonePositionHistory: ArrayList<Int> = ArrayList()
    var scoring: GameScoring = GameScoring(boardSize, winSequenceCount)
    fun exploreMove(player: Int): Float{
        return exploreMove(
            whiteBoard, blackBoard,
            stonePositionHistory,
            player
        )
    }
    fun isGameOver():Boolean{
        if (scoring.gameDraw(whiteBoard,blackBoard)){
            return true
        }
        if (scoring.checkWin(whiteBoard)){
            return true
        }
        if (scoring.checkWin(blackBoard)){
            return true
        }
        return false
    }

    private fun exploreMove(
        whiteBoard: BooleanArray, blackBoard: BooleanArray, stoneHistory: ArrayList<Int>, player: Int): Float{
        val lastStonePosition = stoneHistory[stoneHistory.size-1]
        if (player == GomokuBoard.PLAYER_WHITE){
            if (scoring.checkLose(blackBoard, whiteBoard, lastStonePosition)){
                return -1.0f
            }
            if (scoring.checkWin(blackBoard)){
                return 1.0f
            }
        } else if (player == GomokuBoard.PLAYER_BLACK){
            if (scoring.checkLose(whiteBoard, blackBoard, lastStonePosition)){
                return 1.0f
            }
            if (scoring.checkWin(whiteBoard)){
                return -1.0f
            }
        }

        if (scoring.gameDraw(blackBoard, whiteBoard)){return 0.0f}
        val gomokuBoard = getBoardState(stoneHistory)
        var nextPosition = gomokuBoard.selectMove(player)
        val nextWhiteBoard = copyBooleanArray(whiteBoard)
        val nextBlackBoard = copyBooleanArray(blackBoard)
        if (nextWhiteBoard[nextPosition]){
            throw RuntimeException("board not empty")
        } else if (nextBlackBoard[nextPosition]){
            throw RuntimeException("board not empty")
        }
        if (player == GomokuBoard.PLAYER_WHITE) {
            nextWhiteBoard[nextPosition] = true
        } else if (player == GomokuBoard.PLAYER_BLACK) {
            nextBlackBoard[nextPosition] = true
        }
        val nextStoneHistory = ArrayList(stoneHistory)
        nextStoneHistory.add(nextPosition)
        var nextExplore = exploreMove(
            nextWhiteBoard, nextBlackBoard,
            nextStoneHistory,
            gomokuBoard.nextPlayer(player))
        //var y = nextPosition/boardSize
        //var x = nextPosition-boardSize*y
//        if (player == GomokuBoard.PLAYER_BLACK){
//            gomokuBoard.putBlackStone(x, y)
//        } else {
//            gomokuBoard.putWhiteStone(x, y)
//        }
        gomokuBoard.notifyReward(nextPosition, nextExplore)
        return nextExplore
    }

    fun copyBooleanArray(board: BooleanArray): BooleanArray{
//        val newArray = BooleanArray(board.size)
//        for (i in 0 until newArray.size){
//            newArray[i] = board[i]
//        }
//        return newArray
        return board.copyOf()
    }

    fun removeStatesWithPreviousHistories(){
        val stateKeys = gameBoardStates.keys
        val removeKeysList = ArrayList<ArrayList<Int>>()
        for (keyHistory in stateKeys){
            if (keyHistory.size < stonePositionHistory.size){
                removeKeysList.add(keyHistory)
            }
        }
        for(keyHistory in removeKeysList){
            gameBoardStates.remove(keyHistory)
        }
    }

    fun cannotPutBlackStone(x: Int, y: Int): Boolean {
        val positionIndex = x + y * boardSize
        val tempBlackBoard = blackBoard.copyOf()
        val tempWhiteBoard = whiteBoard.copyOf()
        tempBlackBoard[positionIndex] = true
        return scoring.checkLose(tempBlackBoard, tempWhiteBoard, positionIndex)
    }

    fun putBlackStone(x: Int, y: Int){
        val positionIndex = x + y * boardSize
        blackBoard[positionIndex] = true
        
        //removeStatesWithPreviousHistories()
        gameBoardStates = HashMap()
        stonePositionHistory.add(positionIndex)
    }
    fun explicitMoveAvailable(): Boolean {
        return (getBoardState(stonePositionHistory).selectExplicitMove(GomokuBoard.PLAYER_WHITE) >= 0)
    }
    fun doMove(): Int {
        // CPU move
        val gomokuBoard = getBoardState(stonePositionHistory)

        val nextPositionIndex = gomokuBoard.selectMove(GomokuBoard.PLAYER_WHITE)
        whiteBoard[nextPositionIndex] = true
        gameBoardStates = HashMap()
        stonePositionHistory.add(nextPositionIndex)
        return nextPositionIndex
    }

    fun getBoardState(stonePositionHistory: ArrayList<Int>): GomokuBoard{
        if (gameBoardStates.containsKey(stonePositionHistory)){
            val cachedState = gameBoardStates[stonePositionHistory]!!
//            if (cachedState.stonePositionHistory.size != stonePositionHistory.size){
//                throw RuntimeException("cache did not match")
//            }
            return cachedState
        }

        val stateBlackBoard = BooleanArray(boardSize *boardSize)
        val stateWhiteBoard = BooleanArray(boardSize * boardSize)
        for(i in 0 until stonePositionHistory.size){
            if (i % 2 == 0){
                stateBlackBoard[stonePositionHistory[i]] = true
            } else {
                stateWhiteBoard[stonePositionHistory[i]] = true
            }
        }

        val newGomokeBoard = GomokuBoard(
            boardSize, winSequenceCount,
            stateBlackBoard, stateWhiteBoard,
            ArrayList(stonePositionHistory)
        )
        gameBoardStates.put(
            stonePositionHistory, newGomokeBoard
        )

        return newGomokeBoard
    }
    fun getStoneStatus(x: Int, y: Int): Int {
        if (whiteBoard[x + y* boardSize]){
            return STONE_WHITE
        }
        if (blackBoard[x + y * boardSize]){
            return STONE_BLACK
        }
        return STONE_EMPTY
    }
    companion object {
        const val STONE_EMPTY = 0
        const val STONE_BLACK = 1
        const val STONE_WHITE = 2
    }
}
