package com.jwechrs.simplegomokugame
import com.jwechrs.mathbridge.Math

//ignored by js --> import android.hardware.camera2.params.BlackLevelPattern
import com.jwechrs.simplegomokugame.GameScoring.Companion.checkConsequtiveStonesInDirection
import com.jwechrs.touchlessgesture.common.TNLog
//ignored by js --> import java.io.BufferedOutputStream
//ignored by js --> import java.lang.RuntimeException
import kotlin.random.Random

class StoneSequence(val startX: Int, val startY: Int,
val directionX: Int, val directionY:Int, var repeatCount: Int){
    fun getPositionIndex(boardSize: Int): Int{
        return startX + boardSize * startY
    }
}

class GomokuBoard (
    val boardSize: Int,
    val winSequenceCount: Int,
    var blackBoard: BooleanArray,
    var whiteBoard: BooleanArray,
    var stonePositionHistory: ArrayList<Int>) {
    //var boardSize: Int = 10
    //var winSequenceCount: Int = 3
    //var whiteBoard: BooleanArray
    //var blackBoard: BooleanArray
    //var stonePositionHistory: ArrayList<IntArray> = ArrayList()
    //var historyInex: Int = 0
    var positionExploredCount: IntArray
    var positionExplored: BooleanArray
    var reward: FloatArray
    var positionEmpty: BooleanArray
    var nearbyOccupied: BooleanArray
    var blackPartialSequence: ArrayList<StoneSequence> = ArrayList()
    var whitePartialSequence: ArrayList<StoneSequence> = ArrayList()

    val directions = ArrayList<IntArray>()


    init {
        //boardSize = size
        //whiteBoard = BooleanArray(boardSize * boardSize)
        //blackBoard = BooleanArray(boardSize * boardSize)
        positionExploredCount = IntArray(boardSize*boardSize)
        positionExplored = BooleanArray(boardSize * boardSize)
        positionEmpty = BooleanArray(boardSize * boardSize)
        nearbyOccupied = BooleanArray(boardSize * boardSize)
        reward = FloatArray(boardSize * boardSize)
        updatePositionEmpty()
        directions.add(intArrayOf(1, 0))
        directions.add(intArrayOf(1, 1))
        directions.add(intArrayOf(0, 1))
        directions.add(intArrayOf(-1,1))
    }
    fun updatePositionEmpty(){
        for(i in 0 until boardSize*boardSize){
            positionEmpty[i] = (!blackBoard[i]) && (!whiteBoard[i])
        }
    }
    fun initBoard() {
        for (i in 0 until boardSize * boardSize) {
            whiteBoard?.let { it[i] = false }
            blackBoard?.let { it[i] = false }
        }
    }

    fun canPutWhiteStone(x: Int, y: Int): Boolean{
        val whiteStatus = getWhiteBoardStatus(x, y)
        val blackStatus = getBlackBoardStatus(x, y)
        if (whiteStatus){
            return false
        }
        if (blackStatus){
            return false
        }
        // TODO: ダメかどうか判定
        return true
    }
    fun canPutBlackStone(x: Int, y: Int): Boolean{
        val whiteStatus = getWhiteBoardStatus(x, y)
        val blackStatus = getBlackBoardStatus(x, y)
        if (whiteStatus){
            return false
        }
        if (blackStatus){
            return false
        }
        // TODO: ダメかどうか判定
        return true
    }

    fun getWhiteBoardStatus(x: Int, y: Int): Boolean{
        if (x < 0 || y < 0 || x >= boardSize || y >= boardSize){
            return false
        }
        return whiteBoard[x + boardSize * y]
    }
    fun getBlackBoardStatus(x: Int, y: Int): Boolean {
        if (x < 0 || y < 0 || x >= boardSize || y >= boardSize){
            return false
        }
        return blackBoard[x + boardSize * y]
    }

    fun updateExploredBooleanArray(){
        for(i in 0 until boardSize * boardSize){
            positionExplored[i] = (positionExploredCount[i] != 0)
        }
    }
    fun updateNearbyOccupied(newStoneX: Int, newStoneY: Int){
        // すでにある石からの距離が2以下のもののみ選ぶ
        // TODO: --> 3にする
        val nearbyDistance = 2
        for(x in -nearbyDistance until nearbyDistance){
            if (newStoneX + x < 0){
                continue
            }
            if (newStoneX + x >= boardSize){
                break
            }
            for(y in -nearbyDistance until nearbyDistance){
                if (newStoneY + y < 0){
                    continue
                }
                if (newStoneY + y >= boardSize){
                    break
                }
                val positionIndex1 = (newStoneX+x) + (newStoneY+y) * boardSize
                if (!nearbyOccupied[positionIndex1]){
                    nearbyOccupied[positionIndex1] = true
                }
            }
        }
    }
    fun updateNearbyOccupied(){
        for(x in 0 until boardSize){
            for(y in 0 until boardSize){
                if (!isPositionEmpty(x + y * boardSize)){
                    updateNearbyOccupied(x, y)
                }
            }
        }
    }
    fun getNumberOfNotExploredPosition(): Int{
        var count = 0
        for(i in 0 until boardSize*boardSize){
            if (!positionEmpty[i]){
                continue
            }
            if (!nearbyOccupied[i]){
                continue
            }
            if (!positionExplored[i]){
                count++
            }
        }
        return count
    }
    fun getUCBArray(player: Int): DoubleArray{
        // WHITEならｈ+/-を逆転する
        val meanReward = DoubleArray(boardSize* boardSize)
        val totalCount = positionExploredCount.sum()
        for(i in 0 until boardSize * boardSize){
            var exploredCount = positionExploredCount[i].toDouble()

            val u = kotlin.math.sqrt(
                2.0 * Math.log(totalCount.toDouble())/exploredCount
            )
            meanReward[i] = reward[i]/positionExploredCount[i] * player + u
        }
        return meanReward
    }
    public fun notifyReward(positionIndex: Int, rewardValue: Float){
        positionExploredCount[positionIndex] ++
        reward[positionIndex] += rewardValue
    }

    private fun selectUCB(player: Int): Int {
        val meanReward = getUCBArray(player)
        //return meanReward.indexOf(meanReward.max()!!)
        var maxRewardValue =  Double.NEGATIVE_INFINITY
        var maxRewardIndex = -1
        for(i in 0 until boardSize*boardSize){
            if (!positionEmpty[i]){
                continue
            }
            if (maxRewardValue <= meanReward[i]){
                maxRewardIndex = i
                maxRewardValue = meanReward[i]
            }
        }
        if (maxRewardIndex != -1){
            TNLog.i(TAG, "reward for this action=${maxRewardValue}")
            return maxRewardIndex
        } else {
            throw RuntimeException("no move possible!")
        }
    }

    fun nextPlayer(player: Int): Int{
        if (player == PLAYER_BLACK)
            return PLAYER_WHITE
        else if (player == PLAYER_WHITE)
            return PLAYER_BLACK
        else
            throw RuntimeException("unknown player")
    }

//    private fun checkConsequtiveStonesInDirection(boardArray: BooleanArray,
//    startX: Int, startY: Int, directionX: Int, directionY: Int, repeat: Int): Boolean{
//        if (repeat == 0){
//            return true
//        }
//        if (startX >= boardSize || startY >= boardSize || startX < 0 ||startY < 0){
//            return false
//        }
//        return if (!boardArray[startX + startY * boardSize]){
//             false
//        } else {
//            checkConsequtiveStonesInDirection(
//                boardArray,
//                startX + directionX,
//                startY + directionY,
//                directionX, directionY,
//                repeat-1
//            )
//        }
//    }


    fun selectMove(player: Int): Int{
        updatePositionEmpty()
        updateExploredBooleanArray()
        updateNearbyOccupied()
        val explicitMove = selectExplicitMove(player)
        if (explicitMove >= 0){
            return explicitMove
        }
        val notExploredCount = getNumberOfNotExploredPosition()
        if (notExploredCount == 0){
            return selectUCB(player)
            // select from UCB
        } else{
            val randomPositionIndex = (0 until notExploredCount).random()
            var notExploredCount = 0
            for(i in 0 until boardSize*boardSize){
                if (!positionEmpty[i]){continue}
                if (!nearbyOccupied[i]) {continue}
                if (!positionExplored[i]){
                    if (notExploredCount == randomPositionIndex){
                        return i
                    }
                    notExploredCount ++
                }
            }
        }
        TNLog.i(TAG, "could not find random position")
        throw RuntimeException("?")
    }

    fun isPositionEmpty(x: Int, y: Int): Boolean{
        return isPositionEmpty(x + y*boardSize)
    }

    fun isPositionEmpty(positionIndex: Int): Boolean {
        return (!whiteBoard[positionIndex]) && (!blackBoard[positionIndex])
    }

    fun selectCandidateInEmptyPosition(x1: Int, y1: Int, x2: Int, y2: Int): Int{
        var candidate1 = -1
        var candidate2 = -1
        if (x1 >= 0 && y1 >= 0 && x1 < boardSize && y1 < boardSize){
            if (isPositionEmpty(x1 + y1 * boardSize)){
                candidate1 = x1 + y1 * boardSize
            }
        }
        if (x2 >= 0 && y2 >= 0 && x2 < boardSize && y2 <boardSize){
            if (isPositionEmpty(x2+ y2 * boardSize)){
                candidate2 = x2 + y2 * boardSize
            }
        }
        return selectCandidate(candidate1, candidate2)
//        if (candidate1 >= 0 && candidate2 >= 0){
//            // 両方できるときはランダムに選択
//            if (Random.nextBoolean()){
//                return candidate1
//            } else {
//                return candidate2
//            }
//        }
//        if (candidate1 >= 0){
//            return candidate1
//        }
//        if (candidate2 >= 0){
//            return candidate2
//        }
        return -1
    }

    fun getBoardArray(player: Int): BooleanArray{
        return if (player == PLAYER_WHITE){
            whiteBoard
        } else if (player == PLAYER_BLACK){
            blackBoard
        } else {
            throw RuntimeException("unknown player")
        }
    }

    fun selectCandidateInBothEmptyPositions(x1: Int, y1: Int, x2: Int, y2: Int): Int{
        // 両方が空いている場合のみ、0.5の確率で片方を選ぶ
        var candidate1: Int = -1
        var candidate2 = -1
        if (x1 >= 0 && y1 >= 0 && x1 < boardSize && y1 < boardSize){
            if (isPositionEmpty(x1 + y1 * boardSize)){
                candidate1 = x1 + y1 * boardSize
            }
        }
        if (x2 >= 0 && y2 >= 0 && x2 < boardSize && y2 <boardSize){
            if (isPositionEmpty(x2+ y2 * boardSize)){
                candidate2 = x2 + y2 * boardSize
            }
        }
        if (candidate1 >= 0 && candidate2 >= 0){
            return if (Random.nextBoolean()){
                candidate1
            } else {
                candidate2
            }
        }
        return -1
    }

    fun selectCandidate(candidate1: Int, candidate2: Int): Int{
        if (candidate1 >= 0 && candidate2 >= 0){
            // 両方できるときはランダムに選択
            if (Random.nextBoolean()){
                return candidate1
            } else {
                return candidate2
            }
        }
        if (candidate1 >= 0){
            return candidate1
        }
        if (candidate2 >= 0){
            return candidate2
        }
        return -1
    }

    fun positionOnBoard(x: Int, y: Int): Boolean{
        return (x >= 0 && y >= 0 && x < boardSize && y < boardSize)
    }

    fun selectSkipCandidateInEmptyPosition(
        boardArray: BooleanArray,
        xn1: Int, yn1: Int, x1: Int, y1: Int, xn2:Int, yn2: Int, x2: Int, y2: Int): Int {
        // (xn1, yn1)に石があって,(x1,y1)が空いているときは(x1,y1)
        var candidate1 = -1
        var candidate2 = -1
        if (positionOnBoard(xn1, yn1) && positionOnBoard(x1, y1)){
            if (boardArray[xn1 + yn1 * boardSize] && isPositionEmpty(x1, y1)){
                candidate1 = x1 + y1 * boardSize
            }
        }
        if (positionOnBoard(xn2, yn2) && positionOnBoard(x2, y2)){
            if (boardArray[xn2 + yn2 * boardSize] && isPositionEmpty(x2, y2)){
                candidate2 = x2 + y2 * boardSize
            }
        }
        return selectCandidate(candidate1, candidate2)
    }
    private fun occupiedOrOutOfBoard(x: Int, y: Int, boardArray: BooleanArray?=null): Boolean{
        if (x < 0 || y < 0 || x >= boardSize || y >= boardSize){
            return true
        }
        boardArray?.let {
            return it[x + y * boardSize]
        }
        return !isPositionEmpty(x, y)
    }

    public fun selectNextReachMove(player: Int): Int{
        // 両端の空いた3つがあるとき、その端のどちらかを選ぶ
        val boardArray = getBoardArray(player)
        val otherBoard = getBoardArray(nextPlayer(player))
        for(x in 0 until boardSize){
            for(y in 0 until boardSize) {
                if (boardArray[x + y * boardSize]){
                    for (direction in directions){
                        if (checkConsequtiveStonesInDirection(
                                boardSize, boardArray,
                                x, y, direction[0], direction[1],
                                winSequenceCount-2,
                                checkBedinnning = true,
                                checkEnd = true,
                                otherPlayerBoard = otherBoard
                            )){
//                            val candidate = selectCandidateInBothEmptyPositions(
//                                x-direction[0], y-direction[1],
//                                x + (winSequenceCount-2) * direction[0],
//                                y + (winSequenceCount-2) * direction[1]
//                            )
//                            if (candidate >= 0){
//                                return candidate
//                            }
                            val left1X = x-direction[0]
                            val left1Y = y - direction[1]
                            val left2X = x-2*direction[0]
                            val left2Y = y-2*direction[1]
                            val right1X = x + 3 * direction[0]
                            val right1Y = y + 3 * direction[1]
                            val right2X = x + 4 * direction[0]
                            val right2Y = y + 4 * direction[1]
                            var candidate1: Int = -1
                            var candidate2: Int = -1
                            if (occupiedOrOutOfBoard(left1X, left1Y) || occupiedOrOutOfBoard(right1X, right1Y)){
                                return -1
                            }
                            if (!occupiedOrOutOfBoard(left1X, left1Y)){
                                if (!occupiedOrOutOfBoard(left2X, left2Y, otherBoard)){
                                    candidate1 = left1X + left1Y * boardSize
                                }
                            }
                            if (!occupiedOrOutOfBoard(right1X, right1Y)){
                                if (!occupiedOrOutOfBoard(right2X, right2Y, otherBoard)){
                                    candidate2 = right1X + right1Y * boardSize
                                }
                            }
                            val candidate = selectCandidate(candidate1, candidate2)
                            if (candidate >= 0){
                                return candidate
                            }
                        }
                    }
//                    if (checkConsequtiveStonesInDirection(
//                            boardSize,
//                            boardArray,
//                            x, y, 1, 0,
//                            winSequenceCount-2,
//                            checkBedinnning = true,
//                            checkEnd = true
//                        )){
//                        val candidate = selectCandidateInBothEmptyPositions(x-1,y, x + winSequenceCount-2, y)
//                        if (candidate >= 0){
//                            return candidate
//                        }
//
//                    }
//                    if (checkConsequtiveStonesInDirection(
//                            boardSize,
//                            boardArray,
//                            x, y, 1, 1,
//                            winSequenceCount-2,
//                            checkBedinnning = true,
//                            checkEnd = true
//                        )){
//
//                        val candidate = selectCandidateInBothEmptyPositions(x-1, y-1, x + winSequenceCount-2, y+ winSequenceCount-2)
//                        if (candidate >= 0){return candidate}
//                    }
//                    if (checkConsequtiveStonesInDirection(
//                            boardSize,
//                            boardArray,
//                            x, y, 0, 1,
//                            winSequenceCount-2,
//                            checkBedinnning = true,
//                            checkEnd = true
//                        )){
//                        val candidate = selectCandidateInBothEmptyPositions(x, y-1, x, y + winSequenceCount-2)
//                        if (candidate >= 0){return candidate}
//                    }
//                    if (checkConsequtiveStonesInDirection(
//                            boardSize,
//                            boardArray,
//                            x, y, -1, 1,
//                            winSequenceCount-2,
//                            checkBedinnning = true,
//                            checkEnd = true
//                        )){
//
//                        val candidate = selectCandidateInBothEmptyPositions(
//                            x-winSequenceCount+2, y + winSequenceCount-2, x+1, y-1)
//                        if (candidate >= 0){
//                            return candidate
//                        }
//                    }
                }
            }
        }
        return -1
    }

    fun selectNextSkipReachMove(player: Int): Int{
        // 両端の空いた１つとばしの3つがあるとき、隙間を選ぶ
        val boardArray = getBoardArray(player)
        val otherPlayerArray = getBoardArray(nextPlayer(player))

        for(x in 0 until boardSize){
            for(y in 0 until boardSize) {
                if (boardArray[x + y * boardSize]){
                    for (direction in directions) {
                        if (checkConsequtiveStonesInDirection(
                                boardSize, boardArray,
                                x, y, direction[0], direction[1],
                                winSequenceCount - 1,
                                checkBedinnning = true,
                                checkEnd = true,
                                missingTorelance = 1,
                                otherPlayerBoard = otherPlayerArray
                            )
                        ) {
                            for (i in 0 until winSequenceCount - 1) {
                                val sequenceX = x + i * direction[0]
                                val sequenceY = y + i * direction[1]
                                val sequencePositionIndex = sequenceX + sequenceY * boardSize
                                if (!boardArray[sequencePositionIndex]) {
                                    return sequencePositionIndex
                                }
                            }
                        }
                    }
                }
            }
        }
        return -1
    }

    fun selectSkipReachMove(player: Int): Int{
        // とび4があるときに5をうつ
        val boardArray = getBoardArray(player)
        val otherBoard = getBoardArray(nextPlayer(player))
        for(x in 0 until boardSize){
            for(y in 0 until boardSize) {
                if (boardArray[x + y * boardSize]){
                    for (direction in directions){
                        if (checkConsequtiveStonesInDirection(
                                boardSize, boardArray, x, y, direction[0], direction[1],
                                winSequenceCount,
                                checkBedinnning = true,checkEnd = true,
                                missingTorelance = 1
                            )){
                            for(i in 0 until winSequenceCount){
                                val sequenceX = x + i * direction[0]
                                val sequenceY = y + i * direction[1]
                                val sequencePositionIndex = sequenceX + sequenceY *boardSize
                                if (!boardArray[sequencePositionIndex]){
                                    if (otherBoard[sequencePositionIndex]){
                                        continue
                                    }
                                    return sequencePositionIndex
                                }
                            }
                        }
                    }
                }
            }
        }
        return -1
    }


    fun selectReadhMove(player: Int): Int {
        val boardArray = if (player == PLAYER_WHITE){
            whiteBoard
        } else if (player == PLAYER_BLACK){
            blackBoard
        } else {
            throw RuntimeException("unknown player")
        }

        for(x in 0 until boardSize){
            for(y in 0 until boardSize) {
                if (boardArray[x + y * boardSize]){
                    for (direction in directions){
                        if (checkConsequtiveStonesInDirection(
                                boardSize, boardArray, x, y, direction[0], direction[1],
                                winSequenceCount-1,
                                checkBedinnning = true,
                                checkEnd = true
                            )){
                            val candidate = selectCandidateInEmptyPosition(
                                x-direction[0], y-direction[1],
                                x +(winSequenceCount-1) * direction[0], y + (winSequenceCount-1) * direction[1]
                            )
                            if (candidate >= 0){
                                return candidate
                            }
                        }
                    }
//                    if (checkConsequtiveStonesInDirection(
//                            boardSize,
//                            boardArray,
//                            x, y, 1, 0,
//                            winSequenceCount-1,
//                            checkBedinnning = true,
//                            checkEnd = true
//                        )){
//                        val candidate = selectCandidateInEmptyPosition(x-1,y, x + winSequenceCount-1, y)
//                        if (candidate >= 0){
//                            return candidate
//                        }
////                        var leftCandidate = -1
////
////                        var rightCandidate = -1
////                        if (x > 0){
////                            val candidatePositionIndex = (x -1) + y * boardSize
////                            if (isPositionEmpty(candidatePositionIndex)){
////                                //return candidatePositionIndex
////                                leftCandidate = candidatePositionIndex
////                            }
////                        }
////                        if (x + winSequenceCount-1 < boardSize){
////                            val candidatePositionIndex = (x + winSequenceCount-1) + y * boardSize
////                            if (isPositionEmpty(candidatePositionIndex)){
////                                //return candidatePositionIndex
////                                rightCandidate = candidatePositionIndex
////                            }
////                        }
////                        if ((leftCandidate>=0) && (rightCandidate>=0)){
////                            if (Random.nextBoolean()){
////                                return leftCandidate
////                            } else {
////                                return rightCandidate
////                            }
////                        } else if (leftCandidate >= 0){
////                            return leftCandidate
////                        } else if (rightCandidate >= 0){
////                            return rightCandidate
////                         }
//                    }
//                    if (checkConsequtiveStonesInDirection(
//                            boardSize,
//                            boardArray,
//                            x, y, 1, 1,
//                            winSequenceCount-1,
//                            checkBedinnning = true,
//                            checkEnd = true
//                        )){
////                        if (x > 0 && y > 0){
////                            val candidate = (x-1) + (y-1)*boardSize
////                            if (isPositionEmpty(candidate)){return candidate}
////                        }
////                        if (x + winSequenceCount-1 < boardSize && y + winSequenceCount-1 < boardSize){
////                            val candidate = (x + winSequenceCount-1) + (y + winSequenceCount-1) * boardSize
////                            if (isPositionEmpty(candidate)){return candidate}
////                        }
//                        val candidate = selectCandidateInEmptyPosition(x-1, y-1, x + winSequenceCount-1, y+ winSequenceCount-1)
//                        if (candidate >= 0){return candidate}
//                    }
//                    if (checkConsequtiveStonesInDirection(
//                            boardSize,
//                            boardArray,
//                            x, y, 0, 1,
//                            winSequenceCount-1,
//                            checkBedinnning = true,
//                            checkEnd = true
//                        )){
////                        if (y > 0){
////                            val candidate = x + (y-1) * boardSize
////                            if (isPositionEmpty(candidate)){
////                                return candidate
////                            }
////                        }
////                        if (y + winSequenceCount-1 < boardSize){
////                            val candidate = x + (y + winSequenceCount-1) * boardSize
////                            if (isPositionEmpty(candidate)){
////                                return candidate
////                            }
////                        }
//                        val candidate = selectCandidateInEmptyPosition(x, y-1, x, y + winSequenceCount-1)
//                        if (candidate >= 0){return candidate}
//                    }
//                    if (checkConsequtiveStonesInDirection(
//                            boardSize,
//                            boardArray,
//                            x, y, -1, 1,
//                            winSequenceCount-1,
//                            checkBedinnning = true,
//                            checkEnd = true
//                        )){
////                        if (x - winSequenceCount+1 >= 0 && y + winSequenceCount-1 < boardSize){
////                            val candidate = (x-winSequenceCount+1) + (y + winSequenceCount-1) * boardSize
////                            if (isPositionEmpty(candidate)){
////                                return candidate
////                            }
////                        }
////                        if (x + 1 < boardSize && y -1 >= 0){
////                            val candidate = (x+1) + (y-1)* boardSize
////                            if (isPositionEmpty(candidate)){
////                                return candidate
////                            }
////                        }
//                        val candidate = selectCandidateInEmptyPosition(x-winSequenceCount+1, y + winSequenceCount-1, x+1, y-1)
//                        if (candidate >= 0){
//                            return candidate
//                        }
//                    }
                }
            }
        }
        return -1
    }
    fun selectExplicitMove(player: Int, checkReverse: Boolean = true): Int {
        // 明示的な手が使えるか調べる
        // なければ　-1　--> MCTSで手を探す

        // winSequenceCount-1ならびをさがす

        val reachMove = selectReadhMove(player)
        if (reachMove >= 0){
            return reachMove
        }
        val skipReachMove = selectSkipReachMove(player)
        if (skipReachMove >= 0){
            return skipReachMove
        }
        // 相手のリーチも調べる
        val reverseReachMove = selectReadhMove(nextPlayer(player))
        if (reverseReachMove >= 0 && checkReverse){
            return reverseReachMove
        }

        val reverseSkipReachMove =selectSkipReachMove(nextPlayer(player))
        if (reverseSkipReachMove >= 0 && checkReverse){
            return reverseSkipReachMove
        }
        val nextReachMove = selectNextReachMove(player)
        if (nextReachMove >= 0){
            return nextReachMove
        }

        val nextSkipReachmove =selectNextSkipReachMove(player)
        if (nextSkipReachmove >= 0){
            return nextSkipReachmove
        }
        val reverseNextReachMove = selectNextReachMove(nextPlayer(player))
        if (reverseNextReachMove >= 0){
            return reverseNextReachMove
        }
        val reverseNextSkipReachMove = selectNextSkipReachMove(nextPlayer(player))
        if (reverseNextSkipReachMove >= 0){
            return reverseNextSkipReachMove
        }
        return -1
    }
    companion object {
        const val PLAYER_BLACK = 1
        const val PLAYER_WHITE = -1
        const val TAG = "GoBoard"
    }
}
